/*

    ZX-SCREEN to PPM converter
    
*/


#include <stdio.h>
#include <strings.h>
#include <stdlib.h>

#define HOBETASIZE  0x11
#define ZXCOLORSIZE 0x1b00
#define ZXMONOSIZE  0x1800
#define ZXTRICOLOR  0x4800
#define ZXMAXSIZE   ZXTRICOLOR+HOBETASIZE

#define ZXATTRSIZE  0x300

#define MAXRGB   255
#define ZXDARK   192
#define ZXMIDDLE 220
#define ZXBRIGHT 250


#define ANYCOLOR  0x100
#define FGDEFAULT 0x00 
#define BGDEFAULT 0x07
#define BRDEFAULT 0x00
#define FLDEFAULT 0x00

#define FGBLUE  1
#define FGRED   2
#define FGGREEN 4
#define BGBLUE  8
#define BGRED   16
#define BGGREEN 32
#define BBRIGHT 64
#define BFLASH  128
#define SFG	0
#define SBG     3
#define SBRIGHT 6
#define SFLASH  7
#define FGMASK  (FGBLUE|FGRED|FGGREEN)
#define BGMASK  (BGBLUE|BGRED|BGGREEN)

#define GETATTR(fl, br, bg, fg) ((fl<<SFLASH)&BFLASH)|((br<<SBRIGHT)&BBRIGHT)|((bg<<SBG)&BGMASK)|(fg&FGMASK)

#define GETFGRED(attr) 	   (attr&BBRIGHT?(attr&FGRED?ZXBRIGHT:0):(attr&FGRED?ZXDARK:0))
#define GETFGGREEN(attr)   (attr&BBRIGHT?(attr&FGGREEN?ZXBRIGHT:0):(attr&FGGREEN?ZXDARK:0))
#define GETFGBLUE(attr)    (attr&BBRIGHT?(attr&FGBLUE?ZXBRIGHT:0):(attr&FGBLUE?ZXDARK:0))
#define GETBGRED(attr) 	   (attr&BBRIGHT?(attr&BGRED?ZXBRIGHT:0):(attr&BGRED?ZXDARK:0))
#define GETBGGREEN(attr)   (attr&BBRIGHT?(attr&BGGREEN?ZXBRIGHT:0):(attr&BGGREEN?ZXDARK:0))
#define GETBGBLUE(attr)    (attr&BBRIGHT?(attr&BGBLUE?ZXBRIGHT:0):(attr&BGBLUE?ZXDARK:0))
 
#define ZXCHUNK(line, col, scl)  (col+((line<<5)&0xe0)+(scl<<8)+((line<<0x08)&0x1800))
#define ZXATTR(line, col)   	((col+(line<<5))|ZXMONOSIZE)


#define ZXLINES	   24
#define ZXSCLINES  8
#define ZXCOLS     32


static int verbose=0;
static int monochrome=0;
static int bright=0;
static int fgcolor=ANYCOLOR;
static int bgcolor=ANYCOLOR;

static FILE *inf, *outf;


int scrtopgm()
{
int tricolor=0;
int x, y, size;
int line, col, scl, w;
unsigned char wbuf[(ZXCOLS*3*8)*4], attr, chunk[3], *zxscr, *zxptr;

    if ((zxptr=zxscr=malloc(ZXMAXSIZE+1))==NULL) {
			fprintf(stderr, "out of memory!\n");
			return 0;
			}
    
    size=fread(zxscr, 1, ZXMAXSIZE+1, inf);

    switch (size) {
	case ZXCOLORSIZE+HOBETASIZE : 
		zxptr+=HOBETASIZE;
	case ZXCOLORSIZE : 
		if (monochrome) goto setcolors;
		break;
		
	case ZXMONOSIZE+HOBETASIZE : 
		zxptr+=HOBETASIZE;
	case ZXMONOSIZE : 
setcolors:
		if (fgcolor==ANYCOLOR) fgcolor=FGDEFAULT;
		if (bgcolor==ANYCOLOR) bgcolor=BGDEFAULT;
		if (bright==ANYCOLOR) bright=BRDEFAULT;
		attr=GETATTR(FLDEFAULT, bright, bgcolor, fgcolor);
		for (x=0; x<ZXATTRSIZE; x++ ) *(zxptr+ZXMONOSIZE+x)=attr;
		break;	
		
	case ZXTRICOLOR+HOBETASIZE :	
		zxptr+=HOBETASIZE;
	case ZXTRICOLOR :
		tricolor=1;
		break;
		
	default: fprintf(stderr, "unknown input file format!\n"); 
		 free(zxscr);
		 return 0;
	}

    sprintf(wbuf, "P6\n%d %d\n%d\n", ZXCOLS*8, ZXLINES*ZXSCLINES, MAXRGB); w=strlen(wbuf);
    
    for (line=0; line<ZXLINES; line++) {
	for (scl=0; scl<ZXSCLINES; scl++) {
	    if (w+ZXCOLS*3*8>sizeof(wbuf)) {
		if (w!=fwrite(wbuf, 1, w, outf)) {
			    fprintf(stderr, "can't write output file!\n");
			    free(zxscr);
			    return 0;
			    }
		w=0;
		}
	    for (col=0; col<ZXCOLS; col++) {
		if (tricolor) {
		    for (y=0; y<3; y++) chunk[y]=*(zxptr+y*ZXMONOSIZE+ZXCHUNK(line, col, scl));
		    for (x=0; x<8; x++) {
			for (y=0; y<3; y++) {
			    wbuf[w++]= chunk[y]&0x80 ? ZXMIDDLE : 0;
			    chunk[y]=chunk[y]<<1;
			    }
			}
		    }
		else {
		    chunk[0]=*(zxptr+ZXCHUNK(line, col, scl));
		    attr=*(zxptr+ZXATTR(line, col));
		    for (x=0; x<8; x++) {
			wbuf[w++]= chunk[0]&0x80 ? GETFGRED(attr) : GETBGRED(attr);
			wbuf[w++]= chunk[0]&0x80 ? GETFGGREEN(attr) : GETBGGREEN(attr);
			wbuf[w++]= chunk[0]&0x80 ? GETFGBLUE(attr) : GETBGBLUE(attr);
			chunk[0]=chunk[0]<<1;
			}
		    }
		}
	    }
	}
    
    if (w!=0) 
	if (w!=fwrite(wbuf, 1, w, outf)) {
		    fprintf(stderr, "can't write output file!\n");
		    free(zxscr);
		    return 0;
		    }
    free(zxscr);
    return 1;
}



void help()
{
    fprintf(stderr, "(L) CopyLeft Kirill Frolov 2:5030/827.2@fidonet.org\n" \
		    "ZX-SCREEN to PPM converter\n" \
		    "USAGE:\n" \
		    "scrtopgm [-option] [-option]... [input file [output file]]\n" \
		    "default input file is stdin, output to stdout\n" \
		    "OPTIONS:\n" \
		    "   -v   be verbose\n" \
		    "   -m   force monochrome zx-screen\n" \
/*		   
		    "   -bg <color> background color (default is gray)\n" \
		    "   -fg <color> foreground color (default is black)\n" \
*/		    
		    "   -h --help show this help\n");
    exit(1);
}

void paramerror()
{
    fprintf(stderr, "command line parameter error\n");
    exit(1);
}

int main(int argc, char *argv[]) 
{
int i;

    outf=stdout;
    inf=stdin;
    i=0;
    
    while (--argc)  
    	if (argv[++i][0]=='-') {
	    switch (argv[i][1]) {
		case 'v' : verbose=1; break;
		case 'm' : monochrome=1; break;
		case 'h' : help();
		case '-' : if (strcmp(&argv[i][2], "help")==0) help();
			    else paramerror();
		default: paramerror();
		}
	    continue;
	    }
	  else {
	      inf=fopen(argv[i],"rb");
	      if (inf==NULL) {
	              fprintf(stderr, "can't open \"%s\"\n", argv[i]);
		      exit(-1);
	      }
	      argc--;
	      break;
	  }
	
    if (argc) {
	outf=fopen(argv[++i], "wb");
	if (outf==NULL) {
		fprintf(stderr, "can't open \"%s\"\n", argv[i]);
		exit(-1);
	    }
	}

    i=scrtopgm(inf, outf, monochrome, 1);
    
    fclose(inf);
    fclose(outf);
    
    return i?0:1;
}	  
	  
	